@@ -4,14 +4,14 @@
 # Written 2003 - 2007 by
 # Kjetil Torgrim Homme <kjetilho+listadmin@ifi.uio.no>
 #
-# 2016: Johnny A. Solbu <johnny@solbu.net>
+# 2016 - 2021: Johnny A. Solbu <johnny@solbu.net>
 #
 # Thank you, Sam Watkins and Bernie Hoeneisen, for contributions and
 # feedback.
 #
 # Released into public domain.
 
-my $version = "2.42";
+my $version = "2.73";
 my $maintainer = "johnny\@solbu.net";
 
 use HTML::TokeParser;
@@ -31,6 +31,13 @@ use IO::Socket::SSL;
 use Net::INET6Glue::INET_is_INET6;
 
 my $rc = $ENV{"HOME"}."/.listadmin.ini";
+unless (-r $rc) {
+    eval(q{
+         use File::BaseDir;
+         $rc = (File::BaseDir::config_files 'listadmin.ini')[0]
+            || (File::BaseDir::config_home 'listadmin.ini')[0];
+         });
+}
 
 sub usage {
     my ($exit_val) = @_;
@@ -49,6 +56,8 @@ Usage: $0 [-f CONFIGFILE] [-t MINUTES] [{-a|-r} FILE] [-l] [LISTNAME]
   --remove-member ADDRESS
                    Remove ADDRESS from member list
   -l               List subscribers
+  --dry-run        Automatically (S)kip all held messages
+  --discard-all    Automatically (D)iscard all held messages
   LISTNAME         Only process lists with name matching LISTNAME.
 
 If options which modify members are used, LISTNAME must match exactly
@@ -58,7 +67,8 @@ _end_
 }
 
 my ($opt_help, $opt_version, $opt_f, $opt_t, $opt_a, $opt_r,
-    @opt_add_member, @opt_remove_member, $opt_l);
+    @opt_add_member, @opt_remove_member, $opt_l, $opt_dryrun,
+    $opt_discardall);
 my $opt_mail = 1;
 
 GetOptions("help|?" => \$opt_help,
@@ -70,6 +80,8 @@ GetOptions("help|?" => \$opt_help,
 	   "r=s" => \$opt_r,
 	   "add-member=s" => \@opt_add_member,
 	   "remove-member=s" => \@opt_remove_member,
+           "dry-run!" => \$opt_dryrun,
+           "discard-all!" => \$opt_discardall,
 	   "l" => \$opt_l)
 	or usage();
 
@@ -341,6 +353,7 @@ sub approve_messages {
     my $dis_from = $config->{"discard_if_from"};
     my $dis_subj = $config->{"discard_if_subject"};
     my $dis_reas = $config->{"discard_if_reason"};
+    my $apr_from = $config->{"approve_if_from"};
 
     $count = keys (%{$info}) - 1;	# subtract 1 for globals
     my $search_pattern = "";
@@ -410,18 +423,38 @@ _end_
 		}
 		$ans = "d";
 	    }
+	    # Whitelist senders
+	    my $match2 = "";
+	    $ans ||= $config->{"action"};
+	    $match2 = "From_allow" if got_match ($from, $apr_from);
+	    $ans ||= "a" if $match2;
+	    if ($ans && $match2) {
+	        if ($match2 eq "From_allow") {
+	            print "Automatically appoved due to matching $match2\n";
+	        }
+	        $ans = "a";
+	    }
 	    my $def = $listdef;
 	    $def = $change->{$id}->[0]
 		    if defined $change->{$id};
 	    my $pr = $prompt;
 	    $pr .= " [" . uc($def) . "]" if $def;
 	    $pr .= " ? ";
-	    $ans ||= prompt ($pr);
-	    $ans = "q" unless defined $ans;
-	    $ans =~ s/^\s+//;
-	    $ans =~ s/\s+$//;
-	    $ans = $def if $ans eq "" && defined $def;
-	    $ans = lc $ans;
+            if ($opt_dryrun) {
+                # In dry-run mode, we (S)kip everything.
+                $ans = "s";
+            } elsif ($opt_discardall) {
+                # In discard-all mode, we (D)iscard everything.
+                $ans = "d";
+            } else {
+                # Otherwise, we prompt the user for action.
+                $ans ||= prompt ($pr);
+                $ans = "q" unless defined $ans;
+                $ans =~ s/^\s+//;
+                $ans =~ s/\s+$//;
+                $ans = $def if $ans eq "" && defined $def;
+                $ans = lc $ans;
+            }
 	    if ($ans eq "q") {
 		last msgloop;
 	    } elsif ($ans eq "s") {
@@ -553,7 +586,7 @@ _end_
 		if ($mime_headers =~ /content-transfer-encoding:\s+(\S+)/i) {
 		    my $cte = $1;
 		    if ($cte =~ /quoted-printable/i) {
-			$text = MIME::QuotedPrint::decode($text);
+			$text = MIME::QuotedPrint::decode(utf8::is_utf8($text) ? Encode::encode_utf8($text) : $text);
 		    } elsif ($cte =~ /base64/i) {
 			# Don't bother with truncated lines.
 			$text =~ s!([A-Za-z0-9/+=]{72,76})!MIME::Base64::decode_base64($1)!ge;
@@ -643,6 +676,9 @@ sub uio_adminurl {
 sub mailman_url {
     my ($list, $pattern, $params, $action) = @_;
 
+    # Calling mailman_url with $pattern often requires the redirects
+    # _NOT_ to be resolved (i.e. for logging)
+    my $nopattern = !( defined($pattern) or $pattern );
     my ($lp, $domain) = split ('@', $list);
 
     $pattern ||= uio_adminurl ($domain);
@@ -659,9 +695,30 @@ sub mailman_url {
 	$url .= "/$action";
     }
     $url .= "?$params" if $params;
+
+    $url = resolve_redirects($url) if $nopattern;
     return $url;
 }
 
+sub resolve_redirects {
+    # Test whether the URL is a redirect and, if so, replace it with
+    # the destination one.
+    my ($url, $resp, $req);
+    $url = shift;
+    $resp = $ua->get($url);
+    # If the URL could not get to the requested URL, there's no point
+    # in trying any further
+    if (! $resp->is_success) {
+	print STDERR "Error: Fetching $url\n";
+	print STDERR $resp->status_line, "\n";
+	return undef;
+    }
+
+    # $req will have the final URI, even if redirects were followed.
+    $req = $resp->request->uri;
+    return $req;
+}
+
 # Returns a ref to a hash with all the information about pending messages
 sub get_list {
     my ($list, $config, $pw) = @_;
@@ -1387,7 +1444,12 @@ sub parse_approval {
 	$subject .= " [unencoded]";
     }
 
-    $data->{$id}->{"subject"} = decode_rfc2047($subject, $config);
+    $subject = decode_rfc2047($subject, $config);
+
+    # Replace line breaks with a single space character.
+    $subject =~ s/[\n\r]/ /g;
+    
+    $data->{$id}->{"subject"} = $subject;
 
     $body .= "\n" unless $body =~ /\n$/;
     $data->{$id}->{"body"} = $body;
@@ -1422,6 +1484,7 @@ sub read_config {
     my %cur = map { $_ => undef; }
 	    qw (not_spam_if_from
 		not_spam_if_subject
+		approve_if_from
 		discard_if_from
 		discard_if_subject
 		discard_if_reason);
